/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
/***************************************************************************
 * nm-openswan.c : GNOME UI dialogs for configuring openswan VPN connections
 *
 * Copyright (C) 2005 David Zeuthen, <davidz@redhat.com>
 * Copyright (C) 2005 - 2008 Dan Williams, <dcbw@redhat.com>
 * Copyright (C) 2010 Avesh Agarwal <avagarwa@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 **************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdlib.h>
#include <ctype.h>
#include <glib/gi18n-lib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <gnome-keyring.h>
#include <gnome-keyring-memory.h>

#define NM_VPN_API_SUBJECT_TO_CHANGE

#include <nm-vpn-plugin-ui-interface.h>
#include <nm-setting-vpn.h>
#include <nm-setting-connection.h>
#include <nm-setting-ip4-config.h>

#include "src/nm-openswan-service.h"
#include "common-gnome/keyring-helpers.h"
#include "nm-openswan.h"
#include "import-file.h"

#define OPENSWAN_PLUGIN_NAME    _("IPsec based VPN")
#define OPENSWAN_PLUGIN_DESC    _("IPsec, IKEv1, IKEv2 based VPN")
#define OPENSWAN_PLUGIN_SERVICE NM_DBUS_SERVICE_OPENSWAN 

#define ENC_TYPE_SECURE 0
#define ENC_TYPE_WEAK   1
#define ENC_TYPE_NONE   2

#define PW_TYPE_SAVE   0
#define PW_TYPE_ASK	   1
#define PW_TYPE_UNUSED 2

/************** plugin class **************/

static void openswan_plugin_ui_interface_init (NMVpnPluginUiInterface *iface_class);

G_DEFINE_TYPE_EXTENDED (OpenswanPluginUi, openswan_plugin_ui, G_TYPE_OBJECT, 0,
						G_IMPLEMENT_INTERFACE (NM_TYPE_VPN_PLUGIN_UI_INTERFACE,
											   openswan_plugin_ui_interface_init))

/************** UI widget class **************/

static void openswan_plugin_ui_widget_interface_init (NMVpnPluginUiWidgetInterface *iface_class);

G_DEFINE_TYPE_EXTENDED (OpenswanPluginUiWidget, openswan_plugin_ui_widget, G_TYPE_OBJECT, 0,
						G_IMPLEMENT_INTERFACE (NM_TYPE_VPN_PLUGIN_UI_WIDGET_INTERFACE,
											   openswan_plugin_ui_widget_interface_init))

#define OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), OPENSWAN_TYPE_PLUGIN_UI_WIDGET, OpenswanPluginUiWidgetPrivate))

typedef struct {
	GladeXML *xml;
	GtkWidget *widget;
	GtkSizeGroup *group;
	gint orig_dpd_timeout;
} OpenswanPluginUiWidgetPrivate;


#define OPENSWAN_PLUGIN_UI_ERROR openswan_plugin_ui_error_quark ()

static GQuark
openswan_plugin_ui_error_quark (void)
{
	static GQuark error_quark = 0;

	if (G_UNLIKELY (error_quark == 0))
		error_quark = g_quark_from_static_string ("openswan-plugin-ui-error-quark");

	return error_quark;
}

/* This should really be standard. */
#define ENUM_ENTRY(NAME, DESC) { NAME, "" #NAME "", DESC }

GType
openswan_plugin_ui_error_get_type (void)
{
	static GType etype = 0;

	if (etype == 0) {
		static const GEnumValue values[] = {
			/* Unknown error. */
			ENUM_ENTRY (OPENSWAN_PLUGIN_UI_ERROR_UNKNOWN, "UnknownError"),
			/* The specified property was invalid. */
			ENUM_ENTRY (OPENSWAN_PLUGIN_UI_ERROR_INVALID_PROPERTY, "InvalidProperty"),
			/* The specified property was missing and is required. */
			ENUM_ENTRY (OPENSWAN_PLUGIN_UI_ERROR_MISSING_PROPERTY, "MissingProperty"),
			/* The connection was missing invalid. */
			ENUM_ENTRY (OPENSWAN_PLUGIN_UI_ERROR_INVALID_CONNECTION, "InvalidConnection"),
			{ 0, 0, 0 }
		};
		etype = g_enum_register_static ("OpenswanPluginUiError", values);
	}
	return etype;
}

/* This function verifies if the input string represents a valid 
 * ipv4 or not, only if the string is a ipv4 address otherwise consider 
 * the string a hostname or ipv6 address and returns true
 */
static gboolean
valid_ip_if_ip(const char *str) 
{

	char *tmp;
	/* nlabels counts number of parts seprated by dots*/
	int nlabels=0, i=0;
	char tmp_str[1024];
	long int tmp_li;
	gboolean all_valid_numbers = TRUE;

	tmp = (char *)str;
	
	/* assumption is that an ip address must have either a dot or a digit */
	while(*tmp) {
	tmp_str[i++] = *tmp;

		/* This means this may be hostname or IPV6 return TRUE*/
		if(*tmp != '.' && !isdigit(*tmp)) {
		return TRUE;
		}

		if (*tmp == '.'  || *(tmp+1) == '\0' ) {

		if(*tmp == '.') {
			tmp_str[i-1]='\0';
		}
		else {
			tmp_str[i] = '\0';
		}
		nlabels++;
			/* check length of this part for valid ip, maximum is 3 (255) */
			/* !strlen(tmp_str) tests if tmp_str does not contain anything for example 10.. 
			 * or similar to this */
			if (strlen(tmp_str) > 3 || !strlen(tmp_str)) {
				/* this means a part of string has all digits 
				 * but is not valid for ipv4 address */
				all_valid_numbers = FALSE;
			}
			else {
				tmp_li = strtol(tmp_str, (char **) NULL, 10);
				if( tmp_li < 0 || tmp_li > 255) {
				all_valid_numbers = FALSE;
				}
			}

		/*proceed to process another part if not reached to end of string*/
		i=0;
		}
	tmp++;
	}

	/* if flow comes here that means all are either digits or dots, now check for validity*/
	/* check if there are not 4 part separetd by dots then it is an invalid an ip address,
	 * otherwise consider it as a hostname and return true.
	 *  last character should not be dot that is invalid ip 
	 */
	if(nlabels != 4 || *(tmp-1) == '.') {
		return FALSE;
	}
	/* Now check if all the parts are valid integers as per ip address */
	else if (all_valid_numbers){
		/* this means it is a valid IP address*/
		return TRUE; 

	}
	else {
		/* This is invalid ip address*/
		return FALSE;
	}
}

static gboolean
check_validity (OpenswanPluginUiWidget *self, GError **error)
{
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	GtkWidget *widget;
	char *str;

	widget = glade_xml_get_widget (priv->xml, "gateway_entry");
	str = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
	if (!str || !strlen (str) || strstr (str, " ") || strstr (str, "\t") || !valid_ip_if_ip(str)) {
		g_set_error (error,
		             OPENSWAN_PLUGIN_UI_ERROR,
		             OPENSWAN_PLUGIN_UI_ERROR_INVALID_PROPERTY,
		             NM_OPENSWAN_RIGHT);
		return FALSE;
	}

	widget = glade_xml_get_widget (priv->xml, "group_entry");
	str = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
	if (!str || !strlen (str)) {
		g_set_error (error,
		             OPENSWAN_PLUGIN_UI_ERROR,
		             OPENSWAN_PLUGIN_UI_ERROR_INVALID_PROPERTY,
		             NM_OPENSWAN_LEFTID);
		return FALSE;
	}

	return TRUE;
}

static void
stuff_changed_cb (GtkWidget *widget, gpointer user_data)
{
	g_signal_emit_by_name (OPENSWAN_PLUGIN_UI_WIDGET (user_data), "changed");
}

static gboolean
fill_vpn_passwords (OpenswanPluginUiWidget *self, NMConnection *connection)
{
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	GtkWidget *widget;
	gboolean success = FALSE;
	char *password = NULL;
	char *group_password = NULL;

	/* Grab secrets from the keyring */
	if (connection) {
		NMSettingConnection *s_con;
		NMSettingVPN *s_vpn;
		const char *tmp;

		s_vpn = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);

		if (nm_connection_get_scope (connection) == NM_CONNECTION_SCOPE_SYSTEM) {
			if (s_vpn) {
				tmp = nm_setting_vpn_get_secret (s_vpn, NM_OPENSWAN_XAUTH_PASSWORD);
				if (tmp)
					password = gnome_keyring_memory_strdup (tmp);

				tmp = nm_setting_vpn_get_secret (s_vpn, NM_OPENSWAN_PSK_VALUE);
				if (tmp)
					group_password = gnome_keyring_memory_strdup (tmp);
			}
		} else {
			s_con = NM_SETTING_CONNECTION (nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION));

			/* Lookup passwords in the keyring, and if they weren't there, try
			 * the connection itself, which is where they'd be right after import.
			 */
			keyring_helpers_get_one_secret (nm_setting_connection_get_uuid (s_con),
			                                OPENSWAN_USER_PASSWORD,
			                                &password,
			                                NULL);
			if (!password)
				password = gnome_keyring_memory_strdup (nm_setting_vpn_get_secret (s_vpn, NM_OPENSWAN_XAUTH_PASSWORD));

			keyring_helpers_get_one_secret (nm_setting_connection_get_uuid (s_con),
			                                OPENSWAN_GROUP_PASSWORD,
			                                &group_password,
			                                NULL);
			if (!group_password)
				group_password = gnome_keyring_memory_strdup (nm_setting_vpn_get_secret (s_vpn, NM_OPENSWAN_PSK_VALUE));
		}
	}

	/* User password */
	widget = glade_xml_get_widget (priv->xml, "user_password_entry");
	if (!widget)
		goto out;
	if (password)
		gtk_entry_set_text (GTK_ENTRY (widget), password);
	gtk_size_group_add_widget (priv->group, GTK_WIDGET (widget));
	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);

	/* Group password */
	widget = glade_xml_get_widget (priv->xml, "group_password_entry");
	if (!widget)
		goto out;
	if (group_password)
		gtk_entry_set_text (GTK_ENTRY (widget), group_password);
	gtk_size_group_add_widget (priv->group, GTK_WIDGET (widget));
	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);

	success = TRUE;

out:
	gnome_keyring_memory_free (password);
	gnome_keyring_memory_free (group_password);

	return success;
}

static void
show_toggled_cb (GtkCheckButton *button, OpenswanPluginUiWidget *self)
{
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	GtkWidget *widget;
	gboolean visible;

	visible = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (button));

	widget = glade_xml_get_widget (priv->xml, "user_password_entry");
	g_assert (widget);
	gtk_entry_set_visibility (GTK_ENTRY (widget), visible);

	widget = glade_xml_get_widget (priv->xml, "group_password_entry");
	g_assert (widget);
	gtk_entry_set_visibility (GTK_ENTRY (widget), visible);
}

static void
pw_type_changed_helper (OpenswanPluginUiWidget *self, GtkWidget *combo)
{
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	const char *entry = NULL;
	GtkWidget *widget;
	GtkTreeModel *model;

	/* If the user chose "Not required", desensitize and clear the correct
	 * password entry.
	 */
	widget = glade_xml_get_widget (priv->xml, "user_pass_type_combo");
	if (combo == widget)
		entry = "user_password_entry";
	else {
		widget = glade_xml_get_widget (priv->xml, "group_pass_type_combo");
		if (combo == widget)
			entry = "group_password_entry";
	}
	if (!entry)
		return;

	widget = glade_xml_get_widget (priv->xml, entry);
	g_assert (widget);

	model = gtk_combo_box_get_model (GTK_COMBO_BOX (combo));
	switch (gtk_combo_box_get_active (GTK_COMBO_BOX (combo))) {
	case PW_TYPE_ASK:
	case PW_TYPE_UNUSED:
		gtk_entry_set_text (GTK_ENTRY (widget), "");
		gtk_widget_set_sensitive (widget, FALSE);
		break;
	default:
		gtk_widget_set_sensitive (widget, TRUE);
		break;
	}
}

static void
pw_type_combo_changed_cb (GtkWidget *combo, gpointer user_data)
{
	OpenswanPluginUiWidget *self = OPENSWAN_PLUGIN_UI_WIDGET (user_data);

	pw_type_changed_helper (self, combo);
	stuff_changed_cb (combo, self);
}

static void
init_one_pw_combo (OpenswanPluginUiWidget *self,
                   NMSettingVPN *s_vpn,
                   const char *combo_name,
                   const char *key,
                   const char *entry_name)
{
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	int active = -1;
	GtkWidget *widget;
	GtkListStore *store;
	GtkTreeIter iter;
	const char *value = NULL;
	guint32 default_idx = 1;

	/* If there's already a password and the password type can't be found in
	 * the VPN settings, default to saving it.  Otherwise, always ask for it.
	 */
	widget = glade_xml_get_widget (priv->xml, entry_name);
	if (widget) {
		const char *tmp;

		tmp = gtk_entry_get_text (GTK_ENTRY (widget));
		if (tmp && strlen (tmp))
			default_idx = 0;
	}

	store = gtk_list_store_new (1, G_TYPE_STRING);
	if (s_vpn)
		value = nm_setting_vpn_get_data_item (s_vpn, key);

	gtk_list_store_append (store, &iter);
	gtk_list_store_set (store, &iter, 0, _("Saved"), -1);
	if ((active < 0) && value) {
		if (!strcmp (value, NM_OPENSWAN_PW_TYPE_SAVE))
			active = 0;
	}

	gtk_list_store_append (store, &iter);
	gtk_list_store_set (store, &iter, 0, _("Always Ask"), -1);
	if ((active < 0) && value) {
		if (!strcmp (value, NM_OPENSWAN_PW_TYPE_ASK))
			active = 1;
	}

	gtk_list_store_append (store, &iter);
	gtk_list_store_set (store, &iter, 0, _("Not Required"), -1);
	if ((active < 0) && value) {
		if (!strcmp (value, NM_OPENSWAN_PW_TYPE_UNUSED))
			active = 2;
	}

	widget = glade_xml_get_widget (priv->xml, combo_name);
	g_assert (widget);
	gtk_combo_box_set_model (GTK_COMBO_BOX (widget), GTK_TREE_MODEL (store));
	g_object_unref (store);
	gtk_combo_box_set_active (GTK_COMBO_BOX (widget), active < 0 ? default_idx : active);
	pw_type_changed_helper (self, widget);

	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (pw_type_combo_changed_cb), self);
}

static gboolean
init_plugin_ui (OpenswanPluginUiWidget *self, NMConnection *connection, GError **error)
{
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	NMSettingVPN *s_vpn;
	GtkWidget *widget;
	const char *value = NULL;

	s_vpn = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);

	priv->group = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);

	widget = glade_xml_get_widget (priv->xml, "gateway_entry");
	g_return_val_if_fail (widget != NULL, FALSE);
	gtk_size_group_add_widget (priv->group, GTK_WIDGET (widget));
	if (s_vpn) {
		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENSWAN_RIGHT);
		if (value && strlen (value))
			gtk_entry_set_text (GTK_ENTRY (widget), value);
	}
	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);

	widget = glade_xml_get_widget (priv->xml, "group_entry");
	g_return_val_if_fail (widget != NULL, FALSE);
	gtk_size_group_add_widget (priv->group, GTK_WIDGET (widget));
	if (s_vpn) {
		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENSWAN_LEFTID);
		if (value && strlen (value))
			gtk_entry_set_text (GTK_ENTRY (widget), value);
	}
	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);

	/* Fill the VPN passwords *before* initializing the PW type combos, since
	 * knowing if there are passwords when initializing the combos is helpful.
	 */
	fill_vpn_passwords (self, connection);

	init_one_pw_combo (self, s_vpn, "user_pass_type_combo",
	                   NM_OPENSWAN_XAUTH_PASSWORD_INPUT_MODES, "user_password_entry");
	init_one_pw_combo (self, s_vpn, "group_pass_type_combo",
	                   NM_OPENSWAN_PSK_INPUT_MODES, "group_password_entry");

	widget = glade_xml_get_widget (priv->xml, "user_entry");
	g_return_val_if_fail (widget != NULL, FALSE);
	gtk_size_group_add_widget (priv->group, GTK_WIDGET (widget));
	if (s_vpn) {
		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENSWAN_LEFTXAUTHUSER);
		if (value && strlen (value))
			gtk_entry_set_text (GTK_ENTRY (widget), value);
	}
	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);

	/* Phase 1 Algorithms: IKE*/
	widget = glade_xml_get_widget (priv->xml, "phase1_entry");
	g_return_val_if_fail (widget != NULL, FALSE);
	gtk_size_group_add_widget (priv->group, GTK_WIDGET (widget));
	if (s_vpn) {
		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENSWAN_IKE);
		if (value && strlen (value))
			gtk_entry_set_text (GTK_ENTRY (widget), value);
	}
	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);

	/* Phase 2 Algorithms: ESP*/
	widget = glade_xml_get_widget (priv->xml, "phase2_entry");
	g_return_val_if_fail (widget != NULL, FALSE);
	gtk_size_group_add_widget (priv->group, GTK_WIDGET (widget));
	if (s_vpn) {
		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENSWAN_ESP);
		if (value && strlen (value))
			gtk_entry_set_text (GTK_ENTRY (widget), value);
	}
	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);

	widget = glade_xml_get_widget (priv->xml, "domain_entry");
	g_return_val_if_fail (widget != NULL, FALSE);
	gtk_size_group_add_widget (priv->group, GTK_WIDGET (widget));
	if (s_vpn) {
		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENSWAN_DOMAIN);
		if (value && strlen (value))
			gtk_entry_set_text (GTK_ENTRY (widget), value);
	}
	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (stuff_changed_cb), self);

	/*widget = glade_xml_get_widget (priv->xml, "disable_dpd_checkbutton");
	g_return_val_if_fail (widget != NULL, FALSE);
	if (s_vpn) {
		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENSWAN_DPDTIMEOUT);
		if (value) {
			long int tmp;

			errno = 0;
			tmp = strtol (value, NULL, 10);
			if (tmp >= 0 && tmp <= G_MAXUINT32 && errno == 0)
				priv->orig_dpd_timeout = (guint32) tmp;

			if (priv->orig_dpd_timeout == 0)
				gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
		}
	}
	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (stuff_changed_cb), self);*/

	widget = glade_xml_get_widget (priv->xml, "show_passwords_checkbutton");
	g_return_val_if_fail (widget != NULL, FALSE);
	g_signal_connect (G_OBJECT (widget), "toggled",
	                  (GCallback) show_toggled_cb,
	                  self);

	return TRUE;
}

static GObject *
get_widget (NMVpnPluginUiWidgetInterface *iface)
{
	OpenswanPluginUiWidget *self = OPENSWAN_PLUGIN_UI_WIDGET (iface);
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);

	return G_OBJECT (priv->widget);
}

static guint32
handle_one_pw_type (NMSettingVPN *s_vpn, GladeXML *xml, const char *name, const char *key)
{
	GtkWidget *widget;
	GtkTreeModel *model;
	guint32 pw_type;

	widget = glade_xml_get_widget (xml, name);
	model = gtk_combo_box_get_model (GTK_COMBO_BOX (widget));

	pw_type = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
	switch (pw_type) {
	case PW_TYPE_SAVE:
		nm_setting_vpn_add_data_item (s_vpn, key, NM_OPENSWAN_PW_TYPE_SAVE);
		break;
	case PW_TYPE_UNUSED:
		nm_setting_vpn_add_data_item (s_vpn, key, NM_OPENSWAN_PW_TYPE_UNUSED);
		break;
	case PW_TYPE_ASK:
	default:
		pw_type = PW_TYPE_ASK;
		nm_setting_vpn_add_data_item (s_vpn, key, NM_OPENSWAN_PW_TYPE_ASK);
		break;
	}

	return pw_type;
}

static gboolean
update_connection (NMVpnPluginUiWidgetInterface *iface,
                   NMConnection *connection,
                   GError **error)
{
	OpenswanPluginUiWidget *self = OPENSWAN_PLUGIN_UI_WIDGET (iface);
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	NMSettingVPN *s_vpn;
	GtkWidget *widget;
	char *str;
	guint32 upw_type, gpw_type;

	if (!check_validity (self, error))
		return FALSE;

	s_vpn = NM_SETTING_VPN (nm_setting_vpn_new ());
	g_object_set (s_vpn, NM_SETTING_VPN_SERVICE_TYPE, NM_DBUS_SERVICE_OPENSWAN, NULL);

	/* Gateway */
	widget = glade_xml_get_widget (priv->xml, "gateway_entry");
	str = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
	if (str && strlen (str))
		nm_setting_vpn_add_data_item (s_vpn, NM_OPENSWAN_RIGHT, str);

	/* Group name */
	widget = glade_xml_get_widget (priv->xml, "group_entry");
	str = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
	if (str && strlen (str))
		nm_setting_vpn_add_data_item (s_vpn, NM_OPENSWAN_LEFTID, str);

	/* User name*/
	widget = glade_xml_get_widget (priv->xml, "user_entry");
	str = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
	if (str && strlen (str))
		nm_setting_vpn_add_data_item (s_vpn, NM_OPENSWAN_LEFTXAUTHUSER, str);
	
	/* Phase 1 Algorithms: ike */
	widget = glade_xml_get_widget (priv->xml, "phase1_entry");
	str = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
	if (str && strlen (str))
		nm_setting_vpn_add_data_item (s_vpn, NM_OPENSWAN_IKE, str);

	/* Phase 2 Algorithms: esp */
	widget = glade_xml_get_widget (priv->xml, "phase2_entry");
	str = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
	if (str && strlen (str))
		nm_setting_vpn_add_data_item (s_vpn, NM_OPENSWAN_ESP, str);

	/* Domain entry */
	widget = glade_xml_get_widget (priv->xml, "domain_entry");
	str = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
	if (str && strlen (str))
		nm_setting_vpn_add_data_item (s_vpn, NM_OPENSWAN_DOMAIN, str);

	//widget = glade_xml_get_widget (priv->xml, "disable_dpd_checkbutton");
	//if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
	//	nm_setting_vpn_add_data_item (s_vpn, NM_OPENSWAN_DPDTIMEOUT, "0");
	//} else {
		/* If DPD was disabled and now the user wishes to enable it, just
		 * don't pass the DPD_IDLE_TIMEOUT option to openswan and thus use the
		 * default DPD idle time.  Otherwise keep the original DPD idle timeout.
		 */
	//	if (priv->orig_dpd_timeout >= 10) {
	//		char *tmp = g_strdup_printf ("%d", priv->orig_dpd_timeout);
	//		nm_setting_vpn_add_data_item (s_vpn, NM_OPENSWAN_DPDTIMEOUT, tmp);
	//		g_free (tmp);
	//	}
	//}

	upw_type = handle_one_pw_type (s_vpn, priv->xml, "user_pass_type_combo", NM_OPENSWAN_XAUTH_PASSWORD_INPUT_MODES);
	gpw_type = handle_one_pw_type (s_vpn, priv->xml, "group_pass_type_combo", NM_OPENSWAN_PSK_INPUT_MODES);

	/* System secrets get stored in the connection, user secrets are saved
	 * via the save_secrets() hook.
	 */
	if (nm_connection_get_scope (connection) == NM_CONNECTION_SCOPE_SYSTEM) {
		/* User password */
		widget = glade_xml_get_widget (priv->xml, "user_password_entry");
		str = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
		if (str && strlen (str) && (upw_type != PW_TYPE_UNUSED))
			nm_setting_vpn_add_secret (s_vpn, NM_OPENSWAN_XAUTH_PASSWORD, str);

		/* Group password */
		widget = glade_xml_get_widget (priv->xml, "group_password_entry");
		str = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
		if (str && strlen (str) && (gpw_type != PW_TYPE_UNUSED))
			nm_setting_vpn_add_secret (s_vpn, NM_OPENSWAN_PSK_VALUE, str);
	}

	nm_connection_add_setting (connection, NM_SETTING (s_vpn));
	return TRUE;
}

static void
save_one_password (GladeXML *xml,
                   const char *keyring_tag,
                   const char *uuid,
                   const char *id,
                   const char *entry,
                   const char *combo,
                   const char *desc)
{
	GnomeKeyringResult ret;
	GtkWidget *widget;
	const char *password;
	GtkTreeModel *model;
	gboolean saved = FALSE;

	widget = glade_xml_get_widget (xml, combo);
	g_assert (widget);
	model = gtk_combo_box_get_model (GTK_COMBO_BOX (widget));
	if (gtk_combo_box_get_active (GTK_COMBO_BOX (widget)) == PW_TYPE_SAVE) {
		widget = glade_xml_get_widget (xml, entry);
		g_assert (widget);
		password = gtk_entry_get_text (GTK_ENTRY (widget));
		if (password && strlen (password)) {
			ret = keyring_helpers_save_secret (uuid, id, NULL, keyring_tag, password);
			if (ret == GNOME_KEYRING_RESULT_OK)
				saved = TRUE;
			else
				g_warning ("%s: failed to save %s to keyring.", __func__, desc);
		}
	}

	if (!saved)
		keyring_helpers_delete_secret (uuid, keyring_tag);
}

static gboolean
save_secrets (NMVpnPluginUiWidgetInterface *iface,
              NMConnection *connection,
              GError **error)
{
	OpenswanPluginUiWidget *self = OPENSWAN_PLUGIN_UI_WIDGET (iface);
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	NMSettingConnection *s_con;
	const char *id, *uuid;

	s_con = (NMSettingConnection *) nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION);
	if (!s_con) {
		g_set_error (error,
		             OPENSWAN_PLUGIN_UI_ERROR,
		             OPENSWAN_PLUGIN_UI_ERROR_INVALID_CONNECTION,
		             "missing 'connection' setting");
		return FALSE;
	}

	id = nm_setting_connection_get_id (s_con);
	uuid = nm_setting_connection_get_uuid (s_con);

	save_one_password (priv->xml, OPENSWAN_USER_PASSWORD, uuid, id,
	                   "user_password_entry", "user_pass_type_combo", "user password");
	save_one_password (priv->xml, OPENSWAN_GROUP_PASSWORD, uuid, id,
	                   "group_password_entry", "group_pass_type_combo", "group password");

	return TRUE;
}

static NMVpnPluginUiWidgetInterface *
nm_vpn_plugin_ui_widget_interface_new (NMConnection *connection, GError **error)
{
	NMVpnPluginUiWidgetInterface *object;
	OpenswanPluginUiWidgetPrivate *priv;
	char *glade_file;

	if (error)
		g_return_val_if_fail (*error == NULL, NULL);

	object = NM_VPN_PLUGIN_UI_WIDGET_INTERFACE (g_object_new (OPENSWAN_TYPE_PLUGIN_UI_WIDGET, NULL));
	if (!object) {
		g_set_error (error, OPENSWAN_PLUGIN_UI_ERROR, 0, "could not create openswan object");
		return NULL;
	}

	priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (object);

	glade_file = g_strdup_printf ("%s/%s", GLADEDIR, "nm-openswan-dialog.glade");
	priv->xml = glade_xml_new (glade_file, "openswan-vbox", GETTEXT_PACKAGE);
	if (priv->xml == NULL) {
		g_set_error (error, OPENSWAN_PLUGIN_UI_ERROR, 0,
		             "could not load required resources at %s", glade_file);
		g_free (glade_file);
		g_object_unref (object);
		return NULL;
	}
	g_free (glade_file);

	priv->widget = glade_xml_get_widget (priv->xml, "openswan-vbox");
	if (!priv->widget) {
		g_set_error (error, OPENSWAN_PLUGIN_UI_ERROR, 0, "could not load UI widget");
		g_object_unref (object);
		return NULL;
	}
	g_object_ref_sink (priv->widget);

	if (!init_plugin_ui (OPENSWAN_PLUGIN_UI_WIDGET (object), connection, error)) {
		g_object_unref (object);
		return NULL;
	}

	return object;
}

static void
dispose (GObject *object)
{
	OpenswanPluginUiWidget *plugin = OPENSWAN_PLUGIN_UI_WIDGET (object);
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (plugin);

	if (priv->group)
		g_object_unref (priv->group);

	if (priv->widget)
		g_object_unref (priv->widget);

	if (priv->xml)
		g_object_unref (priv->xml);

	G_OBJECT_CLASS (openswan_plugin_ui_widget_parent_class)->dispose (object);
}

static void
openswan_plugin_ui_widget_class_init (OpenswanPluginUiWidgetClass *req_class)
{
	GObjectClass *object_class = G_OBJECT_CLASS (req_class);

	g_type_class_add_private (req_class, sizeof (OpenswanPluginUiWidgetPrivate));

	object_class->dispose = dispose;
}

static void
openswan_plugin_ui_widget_init (OpenswanPluginUiWidget *plugin)
{
}

static void
openswan_plugin_ui_widget_interface_init (NMVpnPluginUiWidgetInterface *iface_class)
{
	/* interface implementation */
	iface_class->get_widget = get_widget;
	iface_class->update_connection = update_connection;
	iface_class->save_secrets = save_secrets;
}

static guint32
get_capabilities (NMVpnPluginUiInterface *iface)
{
	return (NM_VPN_PLUGIN_UI_CAPABILITY_IMPORT | NM_VPN_PLUGIN_UI_CAPABILITY_EXPORT);
}

static NMConnection *
import_from_file (NMVpnPluginUiInterface *iface, const char *path, GError **error)
{
	NMConnection *connection;
	NMSettingConnection *s_con;
	NMSettingVPN *s_vpn;
	GHashTable *import;
	const char *buf;
	gboolean bool_value;
	NMSettingIP4Config *s_ip4;
	
	import = import_file_load (path);
	if (!import) {
		g_set_error (error, 0, 0, "does not look like a %s openswan connection (parse failed)",
		             OPENSWAN_PLUGIN_NAME);
		return NULL;
	}

	connection = nm_connection_new ();
	s_con = NM_SETTING_CONNECTION (nm_setting_connection_new ());
	nm_connection_add_setting (connection, NM_SETTING (s_con));

	s_vpn = NM_SETTING_VPN (nm_setting_vpn_new ());
	g_object_set (s_vpn, NM_SETTING_VPN_SERVICE_TYPE, NM_DBUS_SERVICE_OPENSWAN, NULL);
	nm_connection_add_setting (connection, NM_SETTING (s_vpn));

	s_ip4 = NM_SETTING_IP4_CONFIG (nm_setting_ip4_config_new ());
	nm_connection_add_setting (connection, NM_SETTING (s_ip4));

	/* Connection name */
	if (import_file_lookup_string (import, "conn", "conn", &buf))
		g_object_set (s_con, NM_SETTING_CONNECTION_ID, buf, NULL);
	else {
		g_set_error (error, 0, 0, "connection name error for %s VPN connection (parse failed)",
			OPENSWAN_PLUGIN_NAME);
		g_object_unref (connection);
		return NULL;
        }

	/* Gateway */
	if (import_file_lookup_string (import, "conn", "right", &buf))
		nm_setting_vpn_add_data_item (s_vpn, NM_OPENSWAN_RIGHT, buf);
	else {
		g_set_error (error, 0, 0, "does not look like a %s openswan connection (no gateway specified)",
		             OPENSWAN_PLUGIN_NAME);
		g_object_unref (connection);
		return NULL;
	}

	/* Group name */
	if (import_file_lookup_string (import, "conn", "leftid", &buf)) {
		if(*buf == '@') {
		nm_setting_vpn_add_data_item (s_vpn, NM_OPENSWAN_LEFTID, buf+1);
		}
		else {
		nm_setting_vpn_add_data_item (s_vpn, NM_OPENSWAN_LEFTID, buf);
		}
	}
	else {
		g_set_error (error, 0, 0, "does not look like a %s openswan connection (no GroupName)",
		             OPENSWAN_PLUGIN_NAME);
		g_object_unref (connection);
		return NULL;
	}


	/* Optional settings */

	/* aggressive mode*/
	if (import_file_lookup_bool (import, "conn", "aggrmode", &bool_value)) {
		if (!bool_value) {
			g_set_error (error, 0, 0, "(aggrmode=yes) must be specified for %s openswan connection",
		             OPENSWAN_PLUGIN_NAME);
			g_object_unref (connection);
			return NULL;
		}
	}

	/* authentication method */
	if (import_file_lookup_string (import, "conn", "authby", &buf)) {
		if(strcmp(buf, "secret")) {
			g_set_error (error, 0, 0, "only authby=secret is supported for %s openswan connection",
				         OPENSWAN_PLUGIN_NAME);
			g_object_unref (connection);
			return NULL;
		}

	}

	/* xauth client mode */
	if (import_file_lookup_bool (import, "conn", "leftxauthclient", &bool_value)) {
		if (!bool_value) {
			g_set_error (error, 0, 0, "(leftxauthclient=yes) must be specified for %s openswan connection",
		             OPENSWAN_PLUGIN_NAME);
			g_object_unref (connection);
			return NULL;
		}
	}

	/* mode cfg client mode */
	if (import_file_lookup_bool (import, "conn", "leftmodecfgclient", &bool_value)) {
		if (!bool_value) {
			g_set_error (error, 0, 0, "(leftmodecfgclient=yes) must be specified for %s openswan connection",
		             OPENSWAN_PLUGIN_NAME);
			g_object_unref (connection);
			return NULL;
		}
	}


	/* remote peer type */
	if (import_file_lookup_string (import, "conn", "remote_peer_type", &buf)) {
		if(strcmp(buf, "cisco")) {
			g_set_error (error, 0, 0, "remote_peer_type must only be of type cisco is supported for %s openswan connection",
				         OPENSWAN_PLUGIN_NAME);
			g_object_unref (connection);
			return NULL;
		}

	}

	/* xauth user name*/
	if (import_file_lookup_string (import, "conn", "leftxauthusername", &buf)) {
		nm_setting_vpn_add_data_item (s_vpn, NM_OPENSWAN_LEFTXAUTHUSER, buf);
	}

	/* xauth server mode */
	if (import_file_lookup_bool (import, "conn", "rightxauthserver", &bool_value)) {
		if (!bool_value) {
			g_set_error (error, 0, 0, "(rightxauthserver=yes) should be specified for %s openswan connection",
		             OPENSWAN_PLUGIN_NAME);
			g_object_unref (connection);
			return NULL;
		}
	}

	/* mode cfg server mode */
	if (import_file_lookup_bool (import, "conn", "rightmodecfgserver", &bool_value)) {
		if (!bool_value) {
			g_set_error (error, 0, 0, "(rightmodecfgserver=yes) should be specified for %s openswan connection",
		             OPENSWAN_PLUGIN_NAME);
			g_object_unref (connection);
			return NULL;
		}
	}

	/* network manager type connection*/
	if (import_file_lookup_bool (import, "conn", "nm_configured", &bool_value)) {
		if (!bool_value) {
			g_set_error (error, 0, 0, "(nm_configured=yes) should be specified for %s openswan connection",
		             OPENSWAN_PLUGIN_NAME);
			g_object_unref (connection);
			return NULL;
		}
	}

	/* IKE algorithms*/
	if (import_file_lookup_string (import, "conn", "ike", &buf)) {
		nm_setting_vpn_add_data_item (s_vpn, NM_OPENSWAN_IKE, buf);
	}

	/* ESP algorithms*/
	if (import_file_lookup_string (import, "conn", "esp", &buf)) {
		nm_setting_vpn_add_data_item (s_vpn, NM_OPENSWAN_ESP, buf);
	}


	/* Connection name */
	/*if (import_file_lookup_string (import, "main", "Description", &buf))
		g_object_set (s_con, NM_SETTING_CONNECTION_ID, buf, NULL);*/

	g_hash_table_destroy (import);

	return connection;
}

static gboolean
export_to_file (NMVpnPluginUiInterface *iface,
        const char *path,
        NMConnection *connection,
        GError **error)
{

	NMSettingConnection *s_con;
	NMSettingVPN *s_vpn;
	FILE *f;
	const char *value=NULL;
	const char *default_username=NULL;
	const char *groupname=NULL;
	const char *gateway=NULL;
	const char *phase1_alg_str=NULL;
	const char *phase2_alg_str=NULL;
	gboolean success = FALSE;

	s_con = NM_SETTING_CONNECTION (nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION));	
	s_vpn = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);

	f = fopen (path, "w");
	if (!f) {
		g_set_error (error, 0, 0, "could not open file for writing");
		return FALSE;
	}

	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENSWAN_RIGHT);
	if (value && strlen (value))
		gateway = value;
	else {
		g_set_error (error, 0, 0, "Openswan connection is missing gateway");
		goto done;
	}

	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENSWAN_LEFTID);
	if (value && strlen (value))
		groupname = value;
	else {
		g_set_error (error, 0, 0, "Openswan connection is missing group name");
		goto done;
	}

	default_username = nm_setting_vpn_get_data_item (s_vpn, NM_OPENSWAN_LEFTXAUTHUSER);    
	if ( !default_username || !strlen(default_username)) {
		default_username = nm_setting_vpn_get_user_name (s_vpn);
	}

	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENSWAN_IKE);
	if(value && strlen (value)) {       
		phase1_alg_str=value;
	}

	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENSWAN_ESP);
	if(value && strlen (value)) {
		phase2_alg_str=value;
	}

	fprintf(f,
			"conn %s\n"
			" aggrmode=yes\n"
			" authby=secret\n"
			" left=%%defaultroute\n"
			" leftid=@%s\n"
			" leftxauthclient=yes\n"
			" leftmodecfgclient=yes\n"
			" leftxauthusername=%s\n"
			" right=%s\n"
			" remote_peer_type=cisco\n"
			" rightxauthserver=yes\n"
			" rightmodecfgserver=yes\n"
			" nm_configured=yes\n"
			" rekey=no\n"
			" salifetime=24h\n"
			" ikelifetime=24h\n"
			" keyingtries=1\n"
			" ike=%s\n"
			" esp=%s\n"
			" auto=add\n",
			nm_setting_connection_get_id (s_con),
			groupname ? groupname: "",
			default_username ? default_username : "",
			gateway ? gateway : "",
			phase1_alg_str ? phase1_alg_str : "aes-sha1",
			phase2_alg_str ? phase2_alg_str : "aes-sha1;modp1024");
     
	success = TRUE;
done:
        fclose (f);
        return success;

}

static gboolean
delete_connection (NMVpnPluginUiInterface *iface,
                   NMConnection *connection,
                   GError **error)
{
	NMSettingConnection *s_con;
	const char *id, *uuid;

	/* Remove any secrets in the keyring associated with this connection's UUID */
	s_con = (NMSettingConnection *) nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION);
	if (!s_con) {
		g_set_error (error,
		             OPENSWAN_PLUGIN_UI_ERROR,
		             OPENSWAN_PLUGIN_UI_ERROR_INVALID_CONNECTION,
		             "missing 'connection' setting");
		return FALSE;
	}

	id = nm_setting_connection_get_id (s_con);
	uuid = nm_setting_connection_get_uuid (s_con);

	if (!keyring_helpers_delete_secret (uuid, OPENSWAN_USER_PASSWORD))
		g_message ("%s: couldn't delete user password for '%s'", __func__, id);

	if (!keyring_helpers_delete_secret (uuid, OPENSWAN_GROUP_PASSWORD))
		g_message ("%s: couldn't delete group password for '%s'", __func__, id);

	return TRUE;
}

static NMVpnPluginUiWidgetInterface *
ui_factory (NMVpnPluginUiInterface *iface, NMConnection *connection, GError **error)
{
	return nm_vpn_plugin_ui_widget_interface_new (connection, error);
}

static void
get_property (GObject *object, guint prop_id,
			  GValue *value, GParamSpec *pspec)
{
	switch (prop_id) {
	case NM_VPN_PLUGIN_UI_INTERFACE_PROP_NAME:
		g_value_set_string (value, OPENSWAN_PLUGIN_NAME);
		break;
	case NM_VPN_PLUGIN_UI_INTERFACE_PROP_DESC:
		g_value_set_string (value, OPENSWAN_PLUGIN_DESC);
		break;
	case NM_VPN_PLUGIN_UI_INTERFACE_PROP_SERVICE:
		g_value_set_string (value, OPENSWAN_PLUGIN_SERVICE);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
openswan_plugin_ui_class_init (OpenswanPluginUiClass *req_class)
{
	GObjectClass *object_class = G_OBJECT_CLASS (req_class);

	object_class->get_property = get_property;

	g_object_class_override_property (object_class,
									  NM_VPN_PLUGIN_UI_INTERFACE_PROP_NAME,
									  NM_VPN_PLUGIN_UI_INTERFACE_NAME);

	g_object_class_override_property (object_class,
									  NM_VPN_PLUGIN_UI_INTERFACE_PROP_DESC,
									  NM_VPN_PLUGIN_UI_INTERFACE_DESC);

	g_object_class_override_property (object_class,
									  NM_VPN_PLUGIN_UI_INTERFACE_PROP_SERVICE,
									  NM_VPN_PLUGIN_UI_INTERFACE_SERVICE);
}

static void
openswan_plugin_ui_init (OpenswanPluginUi *plugin)
{
}

static void
openswan_plugin_ui_interface_init (NMVpnPluginUiInterface *iface_class)
{
	/* interface implementation */
	iface_class->ui_factory = ui_factory;
	iface_class->get_capabilities = get_capabilities;
	iface_class->import_from_file = import_from_file;
	iface_class->export_to_file = export_to_file;
	iface_class->get_suggested_name = NULL;
	iface_class->delete_connection = delete_connection;
}


G_MODULE_EXPORT NMVpnPluginUiInterface *
nm_vpn_plugin_ui_factory (GError **error)
{
	if (error)
		g_return_val_if_fail (*error == NULL, NULL);

	return NM_VPN_PLUGIN_UI_INTERFACE (g_object_new (OPENSWAN_TYPE_PLUGIN_UI, NULL));
}

